#include <ArduinoBLE.h>  // Arduino BLE library
#include <LSM6DS3.h>
#include <Wire.h>
#include <nRF52840.h>

#include <nrfx_saadc.h>
#include <AnalogIn.h>
#include <pinDefinitions.h>

#define CLEAR_STEP true
#define NOT_CLEAR_STEP false


//Create an instance of class LSM6DS3
LSM6DS3 xIMU(I2C_MODE, 0x6A);       //I2C device address 0x6A
LSM6DS3 pedometer(I2C_MODE, 0x6A);  //I2C device address 0x6A

/* Online GUID / UUID Generator:
https://www.guidgenerator.com/online-guid-generator.aspx
64cf715d-f89e-4ec0-b5c5-d10ad9b53bf2
*/
// Nilai ambang untuk klasifikasi gerakan
const float DIAM_THRESHOLD = 0.7;
const float WALK_THRESHOLD = 1;   // Ambang batas akselerasi untuk berjalan
const float BRISK_THRESHOLD = 1.2;  // Ambang batas akselerasi untuk berjalan cepat

// Status gerakan saat ini
int64_t movementStatus = 0;  // 0: diam, 1: berjalan, 2: jalan cepat, 3: berlari
int64_t stepCount1 = 0;
float distance1 = 0;
//float totalDistance = 0;
//float movementStatus = 0;

// UUid for Service
const char* UUID_serv = "64cf715d-f89e-4ec0-b5c5-d10ad9b53bf2";

// UUids for accelerometer characteristics
const char* UUID_ax = "64cf715e-f89e-4ec0-b5c5-d10ad9b53bf2";
const char* UUID_ay = "64cf715f-f89e-4ec0-b5c5-d10ad9b53bf2";
const char* UUID_az = "64cf7160-f89e-4ec0-b5c5-d10ad9b53bf2";
const char* UUID_stepCount = "64cf7161-f89e-4ec0-b5c5-d10ad9b53bf2";
const char* UUID_distance = "64cf7163-f89e-4ec0-b5c5-d10ad9b53bf2";
const char* UUID_movementStatus = "64cf7162-f89e-4ec0-b5c5-d10ad9b53bf2";

// BLE Service
BLEService myService(UUID_serv);

// BLE Characteristics
BLEFloatCharacteristic chAX(UUID_ax, BLERead | BLENotify);
BLEFloatCharacteristic chAY(UUID_ay, BLERead | BLENotify);
BLEFloatCharacteristic chAZ(UUID_az, BLERead | BLENotify);
BLEFloatCharacteristic chstepCount(UUID_stepCount, BLERead | BLENotify);
BLEFloatCharacteristic chmovementStatus(UUID_movementStatus, BLERead | BLENotify);
BLEFloatCharacteristic chdistance(UUID_distance, BLERead | BLENotify);

//Xiao battery;
// class HackAnalogIn : public mbed::AnalogIn {
//   using mbed::AnalogIn::AnalogIn;
// public:
//   analogin_t getAnalogIn_t();
// };

// analogin_t HackAnalogIn::getAnalogIn_t() {
//   return this->_adc;
// }

// void startReadingBatteryLevel(nrf_saadc_value_t* buffer) {
//   auto pin = PIN_VBAT;
//   PinName name = analogPinToPinName(pin);
//   if (name == NC) { return; }
//   HackAnalogIn* adc = static_cast<HackAnalogIn*>(analogPinToAdcObj(pin));
//   if (adc == NULL) {
//     adc = new HackAnalogIn(name);
//     analogPinToAdcObj(pin) = static_cast<mbed::AnalogIn*>(adc);
// #ifdef ANALOG_CONFIG
//     if (isAdcConfigChanged) {
//       adc->configure(adcCurrentConfig);
//     }
// #endif
//   }

//   nrfx_saadc_buffer_convert(buffer, 1);
//   nrfx_err_t ret = nrfx_saadc_sample();
//   if (ret == NRFX_ERROR_BUSY) {
//     // failed to start sampling
//     return;
//   }
// }

void setup() {
  Serial.begin(115200);
  Serial.println("Seeed XIAO BLE Sense");

  // bool err = false;

  // //LED
  // pinMode(LEDR, OUTPUT);     // onboard led red
  // pinMode(LEDB, OUTPUT);     // onboard led blue
  // digitalWrite(LEDR, HIGH);  // led red off
  // digitalWrite(LEDB, HIGH);  // led blue off

  // init IMU
  if (xIMU.begin() != 0) {
    Serial.println("Device error");
  } else {
    Serial.println("Device OK!");
  }

  //Configure LSM6DS3 as pedometer
  if (0 != config_pedometer(NOT_CLEAR_STEP)) {
    Serial.println("Configure pedometer fail!");
  }
  Serial.println("Success to Configure pedometer!");

  // init BLE
  if (!BLE.begin()) {
    Serial.println("BLE: failed");
    while (1)
      ;
  }
  Serial.println("BLE: ok");


  //error: flash led forever
  // if (err) {
  //   Serial.println("Init error. System halted");
  //   while (1) {
  //     digitalWrite(LEDR, LOW);
  //     delay(1000);
  //     digitalWrite(LEDR, HIGH);  // led on
  //     delay(1000);
  //   }
  
 
  // }

  // BLE service
  // correct sequence:
  // set BLE name > advertised service > add characteristics > add service > set initial values > advertise

  // Set BLE name
  BLE.setLocalName("XIAO BLE Sense");
  BLE.setDeviceName("XIAO-BLE-Sense");

  // Set advertised Service
  BLE.setAdvertisedService(myService);

  // Add characteristics to the Service
  myService.addCharacteristic(chAX);
  myService.addCharacteristic(chAY);
  myService.addCharacteristic(chAZ);
  myService.addCharacteristic(chstepCount);
  myService.addCharacteristic(chmovementStatus);
  myService.addCharacteristic(chdistance);

  // add service to BLE
  BLE.addService(myService);

  // characteristics initial values
  chAX.writeValue(0);
  chAY.writeValue(0);
  chAZ.writeValue(0);
  chstepCount.writeValue(0);
  chmovementStatus.writeValue(0);
  //chdistance.writeValue(0);

  // start advertising
  BLE.advertise();
  Serial.println("Advertising started");
  Serial.println("Bluetooth device active, waiting for connections...");
}

//batt
// nrf_saadc_value_t BatteryLevel = { 0 };

void loop() {
  static long preMillis = 0;


  // listen for BLE centrals devices
  BLEDevice central = BLE.central();

  // central device connected?
  if (central) {
    digitalWrite(LEDB, LOW);  // turn on the blue led
    Serial.print("Connected to central: ");
    Serial.println(central.address());  // central device MAC address

    // while the central is still connected to peripheral:
    while (central.connected()) {
      long curMillis = millis();
      if (preMillis > curMillis) preMillis = 0;  // millis() rollover?
      if (curMillis - preMillis >= 50)           // check values every 30mS
      {
        preMillis = curMillis;
        updateValues();  // call function for updating value to send to central
        // readBattery();
      }
    }  // still here while central connected
    // central disconnected:
    digitalWrite(LEDB, HIGH);
    Serial.print("Disconnected from central: ");
    Serial.println(central.address());
  } 
  else {
    // updateValues();  // call function for updating value to send to central
    // // readBattery();
    delay(300);
  }
}

void updateValues() {
  uint8_t averages = 10;  // average on this values count (accelerometer)
  uint8_t dataByte = 0;
  uint16_t stepCount = 0;

  pedometer.readRegister(&dataByte, LSM6DS3_ACC_GYRO_STEP_COUNTER_H);
  stepCount = (dataByte << 8) & 0xFFFF;

  pedometer.readRegister(&dataByte, LSM6DS3_ACC_GYRO_STEP_COUNTER_L);
  stepCount |= dataByte;

  // accelerometer averaged values/actual values
  static float ax = 0;
  static float ay = 0;
  static float az = 0;
  //static float stp=0;
  float ax1, ay1, az1;

  static uint8_t i_a = 0;  // accelerometer readings counter

  // read accelerometer values
  i_a++;
  ax1 = xIMU.readFloatAccelX();
  ay1 = xIMU.readFloatAccelY();
  az1 = xIMU.readFloatAccelZ();

  ax += ax1;
  ay += ay1;
  az += az1;
  stepCount += stepCount1;
  // movementStatus += movementStatus1;
  // distance += distance1;
  //menghitung jarak
  distance1 = stepCount * 0.5;
 

  ax /= averages;
  ay /= averages;
  az /= averages;
  //stepCount/=averages;
  Serial.print("Accelerometer: " + String(ax) + "," + String(ay) + "," + String(az));
  Serial.print(" || ");
  Serial.print("Step: " + String((int)stepCount));
  Serial.print(" || ");
  Serial.println("movement: " + String((int)movementStatus));
  Serial.print(" || ");
  Serial.println("Distance: " + String(distance1));

  // nrf_saadc_value_t BatteryLevel;  // Declare BatteryLevel here
  // startReadingBatteryLevel(&BatteryLevel);
  // // float vBat = (float)BatteryLevel / 4096 * 3.3 / 510 * (1000 + 510);
  // Serial.print("\nBAT: ");
  // Serial.println(vBat);
  // Serial.println(" \n");

  chAX.writeValue(ax);
  chAY.writeValue(ay);
  chAZ.writeValue(az);
  chstepCount.writeValue(stepCount);
  chmovementStatus.writeValue(movementStatus);
  chdistance.writeValue(distance1);

  ax = 0;
  ay = 0;
  az = 0;
  stepCount1 = 0;
  
  i_a = 0;

  //Hitung magnitudo akselerasi
  float magnitude = sqrt((ax1 * ax1) + (ay1 * ay1) + (az1 * az1));

  // Klasifikasikan gerakan berdasarkan magnitudo akselerasi
  if (magnitude < DIAM_THRESHOLD) {
    movementStatus = 0;  // Diam
  } else if (magnitude > DIAM_THRESHOLD && magnitude <= WALK_THRESHOLD) {
    movementStatus = 1;  // Berjalan
  } else if (magnitude > WALK_THRESHOLD && magnitude <= BRISK_THRESHOLD) {
    movementStatus = 2;  // Jalan Cepat
  } else if (magnitude > BRISK_THRESHOLD) {
    movementStatus = 3;  // Berlari
  }

  // Cetak status gerakan ke monitor serial
  Serial.print("Classifier: ");
  switch (movementStatus) {
    case 0:
      Serial.println("Diam");
      break;
    case 1:
      Serial.println("Berjalan");
      break;
    case 2:
      Serial.println("Jalan Cepat");
      break;
    case 3:
      Serial.println("Berlari");
      break;
  }

  delay(300);
}

//Setup pedometer mode
int config_pedometer(bool clearStep) {
  uint8_t errorAccumulator = 0;
  uint8_t dataToWrite = 0;  //Temporary variable

  //Setup the accelerometer******************************
  dataToWrite = 0;

  //  dataToWrite |= LSM6DS3_ACC_GYRO_BW_XL_200Hz;
  dataToWrite |= LSM6DS3_ACC_GYRO_FS_XL_2g;
  dataToWrite |= LSM6DS3_ACC_GYRO_ODR_XL_26Hz;


  // Step 1: Configure ODR-26Hz and FS-2g
  errorAccumulator += pedometer.writeRegister(LSM6DS3_ACC_GYRO_CTRL1_XL, dataToWrite);

  // Step 2: Set bit Zen_G, Yen_G, Xen_G, FUNC_EN, PEDO_RST_STEP(1 or 0)
  if (clearStep) {
    errorAccumulator += pedometer.writeRegister(LSM6DS3_ACC_GYRO_CTRL10_C, 0x3E);
  } else {
    errorAccumulator += pedometer.writeRegister(LSM6DS3_ACC_GYRO_CTRL10_C, 0x3C);
  }

  // Step 3:	Enable pedometer algorithm
  errorAccumulator += pedometer.writeRegister(LSM6DS3_ACC_GYRO_TAP_CFG1, 0x40);

  //Step 4:	Step Detector interrupt driven to INT1 pin, set bit INT1_FIFO_OVR
  errorAccumulator += pedometer.writeRegister(LSM6DS3_ACC_GYRO_INT1_CTRL, 0x10);

  return errorAccumulator;
}

BLECharacteristic txBatCharacteristic;

// void readBattery() {

//   startReadingBatteryLevel(&BatteryLevel);
//   // check if ADC conversion has completed
//   if (nrf_saadc_event_check(NRF_SAADC_EVENT_DONE)) {
//     // ADC conversion completed. Reading is stored in BatteryLevel
//     nrf_saadc_event_clear(NRF_SAADC_EVENT_DONE);
//     float vBat = (float)BatteryLevel / 4096 * 3.3 / 510 * (1000 + 510);
//     // write value to characteristic or things you want to do

//     // Serial.print(" || ");
//     // Serial.print("Bat: ");
//     // Serial.println(vBat);

//     int intBat = int(vBat);
//     uint8_t byteBat = (uint8_t)intBat;
//     txBatCharacteristic.writeValue(byteBat);

//     String strBat = String(vBat);  // Convert float to String
//     int len = strBat.length();
//     char data[len];
//     strBat.toCharArray(data, len + 1);                    // +1 for null terminator
//     txBatCharacteristic.writeValue((uint8_t*)data, len);  // Cast char array to uint8_t*
//   }
// }